import sys
sys.path.append('versatile_diffusion')
import os
import os.path as osp
import PIL
from PIL import Image
from pathlib import Path
import numpy as np
import numpy.random as npr

import torch
import torchvision.transforms as tvtrans
from lib.cfg_helper import model_cfg_bank
from lib.model_zoo import get_model
from lib.model_zoo.ddim_vd import DDIMSampler_VD
from lib.experiments.sd_default import color_adjust, auto_merge_imlist
from torch.utils.data import DataLoader, Dataset

from lib.model_zoo.vd import VD
from lib.cfg_holder import cfg_unique_holder as cfguh
from lib.cfg_helper import get_command_line_args, cfg_initiates, load_cfg_yaml
import matplotlib.pyplot as plt
from skimage.transform import resize, downscale_local_mean

import argparse
parser = argparse.ArgumentParser(description='Argument Parser')
parser.add_argument("-sub", "--sub",help="Subject Number",default=1)
parser.add_argument("-diff_str", "--diff_str",help="Diffusion Strength",default=0.75)
parser.add_argument("-mix_str", "--mix_str",help="Mixing Strength",default=0.4)
parser.add_argument("-vdvae_images", "--vdvae_images", help="vdvae images")
parser.add_argument("-cliptext_preds", "--cliptext_preds", help="cliptext preds")
parser.add_argument("-clipvision_preds", "--clipvision_preds", help="clipvision preds")
parser.add_argument("-out_path", "--out_path", help="out path")
args = parser.parse_args()
sub=int(args.sub)
vdvae_dir = args.vdvae_images
cliptext_preds = args.cliptext_preds
clipvision_preds = args.clipvision_preds
out_path = args.out_path
assert sub in [1,2,5,7]
strength = float(args.diff_str)
mixing = float(args.mix_str)

def regularize_image(x):
        BICUBIC = PIL.Image.Resampling.BICUBIC
        if isinstance(x, str):
            x = Image.open(x).resize([512, 512], resample=BICUBIC)
            x = tvtrans.ToTensor()(x)
        elif isinstance(x, PIL.Image.Image):
            x = x.resize([512, 512], resample=BICUBIC)
            x = tvtrans.ToTensor()(x)
        elif isinstance(x, np.ndarray):
            x = PIL.Image.fromarray(x).resize([512, 512], resample=BICUBIC)
            x = tvtrans.ToTensor()(x)
        elif isinstance(x, torch.Tensor):
            pass
        else:
            assert False, 'Unknown image type'

        assert (x.shape[1]==512) & (x.shape[2]==512), \
            'Wrong image size'
        return x


def sample_image(vdvae_image, cim, ctx, net, sampler):
    n_samples = 1
    ddim_steps = 50
    ddim_eta = 0
    scale = 7.5
    xtype = 'image'
    ctype = 'prompt'

    zim = Image.open(vdvae_image)
   
    zim = regularize_image(zim)
    zin = zim*2 - 1
    zin = zin.unsqueeze(0).cuda(0).half()

    init_latent = net.autokl_encode(zin)
    
    sampler.make_schedule(ddim_num_steps=ddim_steps, ddim_eta=ddim_eta, verbose=False)
    #strength=0.75
    assert 0. <= strength <= 1., 'can only work with strength in [0.0, 1.0]'
    t_enc = int(strength * ddim_steps)
    device = 'cuda:0'
    z_enc = sampler.stochastic_encode(init_latent, torch.tensor([t_enc]).to(device))
    #z_enc,_ = sampler.encode(init_latent.cuda(1).half(), c.cuda(1).half(), torch.tensor([t_enc]).to(sampler.model.model.diffusion_model.device))

    dummy = ''
    utx = net.clip_encode_text(dummy)
    utx = utx.cuda(1).half()
    
    dummy = torch.zeros((1,3,224,224)).cuda(0)
    uim = net.clip_encode_vision(dummy)
    uim = uim.cuda(1).half()
    
    z_enc = z_enc.cuda(1)

    h, w = 512,512
    shape = [n_samples, 4, h//8, w//8]

    
    #c[:,0] = u[:,0]
    #z_enc = z_enc.cuda(1).half()
    
    sampler.model.model.diffusion_model.device='cuda:1'
    sampler.model.model.diffusion_model.half().cuda(1)
    #mixing = 0.4
    
    z = sampler.decode_dc(
        x_latent=z_enc,
        first_conditioning=[uim, cim],
        second_conditioning=[utx, ctx],
        t_start=t_enc,
        unconditional_guidance_scale=scale,
        xtype='image', 
        first_ctype='vision',
        second_ctype='prompt',
        mixed_ratio=(1-mixing), )
    
    z = z.cuda(0).half()
    x = net.autokl_decode(z)
    color_adj='None'
    #color_adj_to = cin[0]
    color_adj_flag = (color_adj!='none') and (color_adj!='None') and (color_adj is not None)
    color_adj_simple = (color_adj=='Simple') or color_adj=='simple'
    color_adj_keep_ratio = 0.5
    
    if color_adj_flag and (ctype=='vision'):
        x_adj = []
        for xi in x:
            color_adj_f = color_adjust(ref_from=(xi+1)/2, ref_to=color_adj_to)
            xi_adj = color_adj_f((xi+1)/2, keep=color_adj_keep_ratio, simple=color_adj_simple)
            x_adj.append(xi_adj)
        x = x_adj
    else:
        x = torch.clamp((x+1.0)/2.0, min=0.0, max=1.0)
        x = [tvtrans.ToPILImage()(xi) for xi in x]
    
    return x

def main():
    cfgm_name = 'vd_noema'
    sampler = DDIMSampler_VD
    pth = 'versatile_diffusion/pretrained/vd-four-flow-v1-0-fp16-deprecated.pth'
    cfgm = model_cfg_bank()(cfgm_name)
    net = get_model()(cfgm)
    sd = torch.load(pth, map_location='cpu')
    net.load_state_dict(sd, strict=False)    

    net.autokl.half()

    torch.manual_seed(0)

    # Might require editing the GPU assignments due to Memory issues
    net.clip.cuda(0)
    net.autokl.cuda(0)

    #net.model.cuda(1)
    sampler = sampler(net)
    #sampler.model.model.cuda(1)
    #sampler.model.cuda(1)

    pred_text = np.load(cliptext_preds)
    pred_text = torch.tensor(pred_text).half().cuda(1)

    pred_vision = np.load(clipvision_preds)
    pred_vision = torch.tensor(pred_vision).half().cuda(1)

    Path(out_path).mkdir(exist_ok=True, parents=True)

    for im_id in range(len(pred_vision)):
        vdvae_image = os.path.join(vdvae_dir, f'{im_id}.png')
        cim = pred_vision[im_id].unsqueeze(0)
        ctx = pred_text[im_id].unsqueeze(0)
        x = sample_image(vdvae_image, cim, ctx, net, sampler)
        x[0].save(os.path.join(out_path, f'{im_id}.png'))

main()
